@@ -133,7 +133,7 @@ action_list_find_mustfix(
 			alist->nr--;
 			list_move_tail(&aitem->list, &immediate_alist->list);
 			immediate_alist->nr++;
-			/* fall through */
+			fallthrough;
 		case XFS_SCRUB_TYPE_BNOBT:
 		case XFS_SCRUB_TYPE_CNTBT:
 		case XFS_SCRUB_TYPE_REFCNTBT:
@@ -258,7 +258,7 @@ action_list_process(
 		}
 	}
 
-	if (xfs_scrub_excessive_errors(ctx))
+	if (scrub_excessive_errors(ctx))
 		return ECANCELED;
 	return 0;
 }
